/**
 * $Id: sc_list.c,v 1.3 2004/06/21 16:44:17 jklein Exp $
 * $Author: jklein $
 * $Log: sc_list.c,v $
 * Revision 1.3  2004/06/21 16:44:17  jklein
 * Added iterator.
 *
 */


#include <stdio.h>
#include "scale/list.h"
#include "scale/alloc.h"


#define _CLEAN_HEAD_TAIL(__list) \
__list->head->contents = __list->tail->contents = NULL


#define _SWAP_LIST_NODE(__lhs, __rhs)\
{\
	spointer temp = __lhs->ontents;\
	__lhs->contents = __rhs->contents;\
	__rhs->contents = temp;\
} while (0)



void
_sc_dump_list(const char *prompt, slist_t *list)
{
	int i;
	printf("%s {", prompt);
	for (i=0; i<list->count; i++) {
		printf("%s", (char *)s_list_get(list, i));
		if (i != list->count - 1)
			printf(", ");
	}
	printf("}\n");
}


static slist_node_t *
_get_node_at(slist_t *list, size_t idx)
{
	slist_node_t *node;
	size_t i;

	if (list->count <= idx)
		return NULL;

	/**
	 * use firstest way to find the 'idx'.
	 */
	if (list->count - idx > list->count / 2) {
		node = list->head->next;
		i = 0;
		while (i != idx) {
			node = node->next;
			++i;
		}
	}
	else {
		node = list->tail;
		i = list->count;
		while (i != idx) {
			node = node->prev;
			--i;
		}
	}

	return node;
}


/**
 * This returns place to add.
 * returns 'end' if no good place to add.
 */
/* static slist_node_t * */
/* _find_place_to_add(sc_compf_t cmpf, slist_node_t *begin, */
/* 				   slist_node_t *end, spointer conts) */
/* { */
/* 	for (; begin != end; begin=begin->next) { */
/* 		if (cmpf(conts, begin->next->contents) >= 0) */
/* 			break; /\* this is it *\/ */
/* 	} */
/* 	return begin; */
/* } */


slist_t *
s_list_alloc()
{
	slist_t *list;

	list = (slist_t *)s_malloc(sizeof(slist_t));

	list->count = 0;
/*	list->compare = compf;*/
	list->head = (slist_node_t *)s_malloc(sizeof(slist_node_t));
	list->tail = (slist_node_t *)s_malloc(sizeof(slist_node_t));

	list->head->contents = NULL;
	list->head->next = list->tail;
	list->head->prev = list->tail;

	list->tail->contents = NULL;
	list->tail->next = list->tail;
	list->tail->prev = list->head;

	return list;
}


void
s_list_free(slist_t *list)
{
	slist_node_t *node, *next;

	for (node=list->head; node != list->tail; node=next) {
		next = node->next;

		node->contents = node->next = node->prev = NULL;

		s_freen(node, sizeof(slist_node_t));
	}
	/* 'node' must be list->tail. */
	node->contents = node->next = node->prev = NULL;
	s_freen(node, sizeof(slist_node_t));
	s_freen(list, sizeof(slist_t));
}


size_t
s_list_size(slist_t *list)
{
	return list->count;
}


void
s_list_add(slist_t *list, const spointer contents)
{
	slist_node_t *newnode;
	slist_node_t *node;

	newnode = (slist_node_t *)s_malloc(sizeof(slist_node_t));
	newnode->contents = contents;
	newnode->next = newnode->prev = NULL;

	/* find place to insert new node. */
	_CLEAN_HEAD_TAIL(list);
	list->tail->contents = contents;

/* 	node = _find_place_to_add(list->compare, list->head, */
/* 							  list->tail, newnode->contents); */

	node = list->tail->prev;

	/** state of link list.
	 * old : [node::prev]->[node]->[node::next]
	 * new : [node::prev]->[node]->[newnode]->[node::next]
	 */
	newnode->next = node->next;
	newnode->prev = node;
	node->next->prev = newnode;
	node->next = newnode;

	++list->count;
}


spointer
s_list_get(slist_t *list, size_t idx)
{
	slist_node_t *node;
	
	if ((node = _get_node_at(list, idx)) == NULL) {
		return NULL;
	}

	return node->contents;
}


spointer
s_list_remove(slist_t *list, size_t idx)
{
	slist_node_t *node;
	spointer contents;

	if ((node = _get_node_at(list, idx)) == NULL) {
		return NULL;
	}

	contents = node->contents;

	/* state of link list.
	 * old: [node::prev]->[node]->[node::next]
	 * new: [node::prev]->[node::next]
	 */
	node->prev->next = node->next;
	node->next->prev = node->prev;

	s_freen(node, sizeof(slist_node_t));
	--list->count;

	return contents;
}


sint32
s_list_find(slist_t *list, sc_compf_t compare, spointer contents)
{
	slist_node_t *node;
	size_t idx;
	sint8 cmpres;

	_CLEAN_HEAD_TAIL(list);

	list->tail->contents = contents;
	node = list->head->next;
	idx = 0;

	while ((cmpres = compare(contents, node->contents)) != 0) {
		++idx;
		node = node->next;
	}

	if (node != list->tail) 
		return idx;
	else
		return -1;
}


sboolean
s_list_join(slist_t *dst, slist_t *src)
{
	slist_node_t *dstnode, *srcnode;
	slist_node_t *newnode;

	_CLEAN_HEAD_TAIL(dst);
	_CLEAN_HEAD_TAIL(src);

	dstnode = dst->tail->prev;
	srcnode = src->head->next;

	while (srcnode != src->tail) {
		/**
		   old: [dstnode] --> [dstnode::next]
		   new: [dstnode] --> [newnode] --> [dstnode::next]
		*/
		newnode = (slist_node_t *)s_malloc(sizeof(slist_node_t));
		newnode->contents = srcnode->contents;
		newnode->prev = dstnode;
		newnode->next = dstnode->next;

		dstnode->next->prev = newnode;
		dstnode->next = newnode;

		dstnode = dstnode->next;
		srcnode = srcnode->next;
	}

	dst->count += src->count;

	return true;
}



typedef struct {
	slist_node_t *next;
} list_itr_manager_t;


static spointer
_s_list_iterator_next(siterator_t *itr)
{
	list_itr_manager_t *mng;
	slist_t *list;
	spointer contents;

	list = (slist_t *)itr->listdata;
	mng = (list_itr_manager_t *)itr->data;

	if (mng->next == NULL)
		return NULL;

	contents = mng->next->contents;

	mng->next = (list->tail != mng->next->next)?
		mng->next->next:NULL;
	return contents;
}


static void
_s_list_iterator_clear(siterator_t *itr)
{
	list_itr_manager_t *mng;
	slist_t *list;

	list = (slist_t *)itr->listdata;
	mng = (list_itr_manager_t *)itr->data;

	mng->next = list->head->next;
}


static void
_s_list_iterator_free(siterator_t *itr)
{
	s_freen(itr->data, sizeof(list_itr_manager_t));
	s_freen(itr, sizeof(siterator_t));
}


siterator_t *
s_list_iterator(slist_t *list)
{
	siterator_t *itr;

	itr = (siterator_t *)s_malloc(sizeof(siterator_t));
	itr->listdata = list;
	itr->data = (list_itr_manager_t *)s_malloc(sizeof(list_itr_manager_t));
	itr->next = _s_list_iterator_next;
	itr->clear = _s_list_iterator_clear;
	itr->free = _s_list_iterator_free;

	itr->clear(itr);

	return itr;
}
